/*
 * MZViewer - A viewer for MZData format Mass Spectra
 * Copyright (c) 2005 Simon Andrews

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */
package uk.ac.bbsrc.babraham.mzdata.displays;

import java.awt.*;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.text.DecimalFormat;
import java.util.Enumeration;

import javax.swing.JOptionPane;
import javax.swing.JPanel;

import uk.ac.bbsrc.babraham.mzdata.MZSpectrum;
import uk.ac.bbsrc.babraham.mzdata.exceptions.MZSpectrumException;

/**
 * @author andrewss
 *
 */
public class SingleSpectrumDisplayPanel extends JPanel {

	private MZSpectrum spectrum;
	private double peakLabelCutoff;
	private double maxIntensity;
	private double minIntensity;
	private double maxMZ;
	private double minMZ;
	private boolean changingPeakLabelCutoff = false;
	private boolean changingXZoom = false;
	private boolean changingYZoom = false;
	private int dragXstarted = 0;
	private int dragYstarted = 0;
	private int dragXcurrent = 0;
	private int dragYcurrent = 0;

	public SingleSpectrumDisplayPanel(MZSpectrum spectrum) {
		this();
		this.setSpectrum(spectrum);
	}

	public SingleSpectrumDisplayPanel() {
		ClickNDragListener l = new ClickNDragListener(this);
		this.addMouseListener(l);
		this.addMouseMotionListener(l);
	}

	public void setSpectrum(MZSpectrum spectrum) {
		this.spectrum = spectrum;
		if (spectrum != null) {
			maxIntensity = spectrum.getMaxIntensity();
			minIntensity = 0;
			peakLabelCutoff = maxIntensity / 3;
			minMZ = spectrum.getMinMZ();
			// We give ourselves 5% breathing room at the end of the spectrum
			maxMZ = spectrum.getMaxMZ();
			maxMZ += (maxMZ - minMZ) / 20;
		}
		this.repaint();
	}

	public void paint(Graphics g0) {
		super.paint(g0);

		Graphics2D g = (Graphics2D) g0;

		int width = this.getWidth();
		int height = this.getHeight();
		
        g.setColor(Color.WHITE);
        g.fillRect(0,0,width,height);

		g.setColor(Color.BLACK);

		// Draw the graph axes and legends
		g.setFont(new Font("sans-serif", Font.BOLD, 12));
		g.drawLine(50, 30, 50, height - 30);

		g.drawLine(50, height - 30, width - 10, height - 30);
		g.drawString("m/z", 50 + ((width - 60) / 2), height - 3);
		// Draw scale text

		if (spectrum == null) {
			return;
		}

		g.setFont(new Font("sans-serif", Font.BOLD, 12));
        String parentIonString = "";
        if (spectrum.getParentMZ()>0) {
            parentIonString = "   Parent Ion="+spectrum.getParentMZ();
        }
		g.drawString("Spectrum ID " + spectrum.getID() + "   msLevel="
				+ spectrum.getMsLevel()+parentIonString, 5, 15);

		// MZ Scale (x-axis)
		double mzDiff = maxMZ - minMZ;
		int mzBase = (int) mzDiff / 5;
		if (mzBase < 1) {
			mzBase = 1;
		}
		int mzXscale = 0;
		int mzXincrease;
		while (true) {
			if (mzBase > 10) {
				mzBase = mzBase / 10;
				++mzXscale;
			} 
			else {
			    mzXincrease = (int)(Math.round(mzBase) * Math.pow(10, mzXscale));
//				if (mzBase > 5) {
//					mzXincrease = (int) (5 * (Math.pow(10, mzXscale)));
//				} 
//				else {
//					mzXincrease = (int) (Math.pow(10, mzXscale));
//				}
				break;
			}
		}

		int currentX = (int) ((minMZ + mzXincrease) - ((minMZ) % mzXincrease));
		g.setFont(new Font("sans-serif", Font.PLAIN, 10));
		while (currentX <= maxMZ) {
			int xOffset = 50 + (int) ((width - 60) * ((currentX - minMZ) / (maxMZ - minMZ)));
			g.drawLine(xOffset, height - 30, xOffset, height - 25);
			g.drawString("" + currentX, xOffset - 10, height - 15);

			currentX += mzXincrease;
		}

		// Intensity Scale (y-axis)
		double intensityDiff = maxIntensity - minIntensity;
		double intensityBase = intensityDiff / 5;
		DecimalFormat df = new DecimalFormat("0.#E0");
		int intensityYscale = 0;
		double intensityYincrease;
		while (true) {
			if (intensityBase > 10) {
				intensityBase = intensityBase / 10;
				++intensityYscale;
			} 
			else if (intensityBase < 1) {
				intensityBase = intensityBase * 10;
				--intensityYscale;
			} 
			else {
			    intensityYincrease = Math.round(intensityBase)* (Math.pow(10, intensityYscale));
				break;
			}
		}

		double currentY = (minIntensity + intensityYincrease)
				- ((minIntensity) % intensityYincrease);
		g.setFont(new Font("sans-serif", Font.PLAIN, 10));
		while (currentY <= maxIntensity) {
			int yOffset = 30 + (int) ((height - 60) * ((maxIntensity - currentY) / (maxIntensity - minIntensity)));
			g.drawLine(50, yOffset, 45, yOffset);
			// Intensity values are BIG - so they get too long to fit into the
			// available space.
			g.drawString(df.format(currentY), 5, yOffset + 3);

			currentY += intensityYincrease;
		}

		// Shade the display to show zoom selections
		if (changingXZoom) {
			g.setColor(new Color(150, 150, 255));
			if (dragXcurrent >= dragXstarted) {
				g.fillRect(dragXstarted, 30, dragXcurrent - dragXstarted,
						height - 60);
			} else {
				g.fillRect(dragXcurrent, 30, dragXstarted - dragXcurrent,
						height - 60);
			}
		} else if (changingYZoom) {
			g.setColor(new Color(150, 150, 255));
			if (dragYcurrent >= dragYstarted) {
				g.fillRect(50, dragYstarted, width - 60, dragYcurrent
						- dragYstarted);
			} else {
				g.fillRect(50, dragYcurrent, width - 60, dragYstarted
						- dragYcurrent);
			}
		}

		// Draw the glyph for the peak label cutoff
		if (peakLabelCutoff > maxIntensity) {
			peakLabelCutoff = maxIntensity;
		}
		if (peakLabelCutoff < minIntensity) {
			peakLabelCutoff = minIntensity;
		}
		int plcY = (int) (30 + ((height - 60) * ((maxIntensity - peakLabelCutoff) / (maxIntensity - minIntensity))));
		g.setColor(Color.RED);
		g.fillPolygon(new int[] { 50, 43, 43, 50 }, new int[] { plcY, plcY - 7,
				plcY + 7, plcY }, 4);

		// If we're chaning the level cutoff then draw a line across the display
		if (changingPeakLabelCutoff) {
			g.drawLine(50, plcY, width - 10, plcY);
		}

		g.setColor(Color.BLACK);

		if (spectrum.getType().equals("discrete")) {
			paintDiscrete(g);
		} else if (spectrum.getType().equals("continuous")) {
			paintContinuous(g);
		}

	}

	private void paintDiscrete(Graphics2D g) {

		int width = getWidth();
		int height = getHeight();

		Enumeration data = spectrum.getDataEnumeration();

		while (data.hasMoreElements()) {
			// Data comes back as an array of two values, MZ and Intensity
			double[] dataPoint = (double[]) data.nextElement();

			if (dataPoint[0] < minMZ || dataPoint[0] > maxMZ
					|| dataPoint[1] < minIntensity) {
				continue;
			}

			double drawnIntensity = dataPoint[1];
			if (drawnIntensity > maxIntensity) {
				drawnIntensity = maxIntensity;
			}

			int xOffset = (int) ((width - 60) * ((dataPoint[0] - minMZ) / (maxMZ - minMZ)));
			int yOffset = (int) ((height - 60) * ((drawnIntensity - minIntensity) / (maxIntensity - minIntensity)));
			g.drawLine(50 + xOffset, height - 30, 50 + xOffset, height
					- (30 + yOffset));

			if (dataPoint[1] > maxIntensity) {
				g.drawLine(45 + xOffset, 42, 55 + xOffset, 32);
				g.drawLine(45 + xOffset, 47, 55 + xOffset, 37);

			}

			if (dataPoint[1] > peakLabelCutoff) {
				g.setFont(new Font("sans-serif", Font.PLAIN, 10));
				g.drawString("" + (float) dataPoint[0], 51 + xOffset, height
						- (30 + yOffset));
			}
		}
	}

	private void paintContinuous(Graphics2D g) {

		double[] lastPoint = null;
		double[] thisPoint = null;
		double[] nextPoint = null;

		int width = getWidth();
		int height = getHeight();

		Enumeration data = spectrum.getDataEnumeration();

		while (data.hasMoreElements()) {
			// Data comes back as an array of two values, MZ and Intensity
			nextPoint = (double[]) data.nextElement();

            // Skip the first point (just preload the data into nextPoint)
			if (thisPoint != null) {
                
                // If we're not getting our data in ascending order of M/Z
                // we can't draw a continuous spectrum, so warn the user 
                // and convert the spectrum to a discrete spectrum
                if (lastPoint != null && thisPoint[0] < lastPoint[0] ) {
                    
                    JOptionPane.showMessageDialog(this,"Continuous spectrum "+spectrum.getID()+" does not provide its data in m/z order and must be drawn as a discrete spectrum","Error",JOptionPane.ERROR_MESSAGE);
                    try {
                        spectrum.setType("discrete");
                    } catch (MZSpectrumException e) {
                        e.printStackTrace();
                    }
                    this.repaint();
                    return;
                }

                // Skip points outside the current MZ or intensity limits
				if (thisPoint[0] < minMZ || thisPoint[0] > maxMZ
						|| thisPoint[1] < minIntensity) {
					lastPoint = thisPoint;
					thisPoint = nextPoint;
					nextPoint = null;
					continue;
				}

                // Scale the intensity down to the top of the chromatogram
                // if the user has zoomed in and this intensity > maximum
                // visible intensity.
				double thisDrawnIntensity = thisPoint[1];
				if (thisDrawnIntensity > maxIntensity) {
					thisDrawnIntensity = maxIntensity;
				}
                  
				int thisXOffset = (int) ((width - 60) * ((thisPoint[0] - minMZ) / (maxMZ - minMZ)));
				int thisYOffset = (int) ((height - 60) * ((thisDrawnIntensity - minIntensity) / (maxIntensity - minIntensity)));

				double lastDrawnIntensity = 0;
				double lastMZ = minMZ;
				if (lastPoint != null) {
					lastDrawnIntensity = lastPoint[1];
					lastMZ = lastPoint[0];
                    if (lastMZ < minMZ) {
                        lastMZ = minMZ;
                    }
				}
				if (lastDrawnIntensity > maxIntensity) {
					lastDrawnIntensity = maxIntensity;
				}
                           
				int lastXOffset = (int) ((width - 60) * ((lastMZ - minMZ) / (maxMZ - minMZ)));
				int lastYOffset = (int) ((height - 60) * ((lastDrawnIntensity - minIntensity) / (maxIntensity - minIntensity)));

                g.drawLine(50 + lastXOffset, height - (30 + lastYOffset),
						50 + thisXOffset, height - (30 + thisYOffset));

				if (thisPoint[1] > peakLabelCutoff && (lastPoint == null || lastPoint[1]<thisPoint[1]) && (nextPoint == null || nextPoint[1]<thisPoint[1])) {
				    g.setFont(new Font("sans-serif",Font.PLAIN,10));
				    g.drawString(""+(float)thisPoint[0],51+thisXOffset,height-(30+thisYOffset));
				}
			}
			lastPoint = thisPoint;
			thisPoint = nextPoint;
			nextPoint = null;
		}

	}

	private void scaleIntensities() {
		Enumeration data = spectrum.getDataEnumeration();

		double tempMaxIntensity = 0;

		while (data.hasMoreElements()) {
			// Data comes back as an array of two values, MZ and Intensity
			double[] dataPoint = (double[]) data.nextElement();

			if (dataPoint[0] < minMZ || dataPoint[0] > maxMZ) {
				continue;
			}

			if (dataPoint[1] > tempMaxIntensity) {
				tempMaxIntensity = dataPoint[1];
			}
		}

		if (tempMaxIntensity > 0) {
			maxIntensity = tempMaxIntensity;
			minIntensity = 0;
			if (peakLabelCutoff > maxIntensity) {
				peakLabelCutoff = maxIntensity / 3;
			}
		}
	}

	private class ClickNDragListener implements MouseListener,
			MouseMotionListener {

		SingleSpectrumDisplayPanel d;

		public ClickNDragListener(SingleSpectrumDisplayPanel d) {
			this.d = d;
		}

		public void mouseClicked(MouseEvent me) {
			if (me.getClickCount() == 2) {
				// A double click resets the zoom
				maxIntensity = spectrum.getMaxIntensity();
				minIntensity = 0;
				minMZ = spectrum.getMinMZ();
				maxMZ = spectrum.getMaxMZ();
				maxMZ += (maxMZ - minMZ) / 20;
				changingPeakLabelCutoff = false;
				changingXZoom = false;
				changingYZoom = false;
				d.repaint();
			}
		}

		public void mousePressed(MouseEvent me) {
			dragXstarted = me.getX();
			dragXcurrent = me.getX();
			dragYstarted = me.getY();
			dragYcurrent = me.getY();

			// First check if we're in the peak height adjuster
			int plcY = (int) (30 + ((d.getHeight() - 60) * ((maxIntensity - peakLabelCutoff) / (maxIntensity - minIntensity))));

			if (dragXstarted >= 43 && dragXstarted <= 50
					&& dragYstarted >= (plcY - 7) && dragYstarted <= (plcY + 7)) {
				changingPeakLabelCutoff = true;
			} else if (dragXstarted <= 50 && dragYstarted >= 30
					&& dragYstarted <= d.getHeight() - 30) {
				changingYZoom = true;
			} else if (dragXstarted >= 50
					&& dragXstarted <= (d.getWidth() - 10)
					&& dragYstarted >= d.getHeight() - 30) {
				changingXZoom = true;
			}

		}

		public void mouseReleased(MouseEvent arg0) {

			if (changingXZoom) {

				if (dragXstarted == dragXcurrent) {
					changingXZoom = false;
					d.repaint();
					return;
				}
				int minX;
				int maxX;

				if (dragXstarted > dragXcurrent) {
					minX = dragXcurrent;
					maxX = dragXstarted;
				} else {
					minX = dragXstarted;
					maxX = dragXcurrent;
				}

				double tempMinMZ = minMZ
						+ ((maxMZ - minMZ) * (((double) minX - 50) / (d
								.getWidth() - 60)));
				double tempMaxMZ = minMZ
						+ ((maxMZ - minMZ) * (((double) maxX - 50) / (d
								.getWidth() - 60)));

				minMZ = tempMinMZ;
				maxMZ = tempMaxMZ;

				d.scaleIntensities();

			} else if (changingYZoom) {
				if (dragYstarted == dragYcurrent) {
					changingYZoom = false;
					d.repaint();
					return;
				}
				int minY;
				int maxY;

				if (dragYstarted > dragYcurrent) {
					minY = dragYcurrent;
					maxY = dragYstarted;
				} else {
					minY = dragYstarted;
					maxY = dragYcurrent;
				}

				double tempMinIntensity = maxIntensity
						- ((maxIntensity - minIntensity) * (((double) maxY - 30) / (d
								.getHeight() - 60)));
				double tempMaxIntensity = maxIntensity
						- ((maxIntensity - minIntensity) * (((double) minY - 30) / (d
								.getHeight() - 60)));

				minIntensity = tempMinIntensity;
				maxIntensity = tempMaxIntensity;

				if (peakLabelCutoff < minIntensity) {
					peakLabelCutoff = minIntensity;
				}
				if (peakLabelCutoff > maxIntensity) {
					peakLabelCutoff = minIntensity
							+ ((maxIntensity - minIntensity) / 3);
				}

			}

			changingPeakLabelCutoff = false;
			changingXZoom = false;
			changingYZoom = false;
			d.repaint();
		}

		public void mouseEntered(MouseEvent arg0) {

		}

		public void mouseExited(MouseEvent arg0) {
			changingPeakLabelCutoff = false;
			changingXZoom = false;
			changingYZoom = false;
		}

		public void mouseDragged(MouseEvent me) {
			dragXcurrent = me.getX();
			dragYcurrent = me.getY();

			if (dragXcurrent < 50) {
				dragXcurrent = 50;
			}
			if (dragXcurrent > d.getWidth() - 10) {
				dragXcurrent = d.getWidth() - 10;
			}
			if (dragYcurrent < 30) {
				dragYcurrent = 30;
			}
			if (dragYcurrent > d.getHeight() - 30) {
				dragYcurrent = d.getHeight() - 30;
			}

			if (changingPeakLabelCutoff) {
				peakLabelCutoff = maxIntensity
						- ((maxIntensity - minIntensity) * (((double) me.getY() - 30) / ((double) d
								.getHeight() - 60)));
			}
			d.repaint();
		}

		public void mouseMoved(MouseEvent arg0) {

		}

	}

}